import os
from typing import Dict

from boa3 import constants
from boa3.boa3 import Boa3
from boa3.compiler.compiler import Compiler
from boa3.exception import CompilerError
from boa3.exception.NotLoadedException import NotLoadedException
from boa3.model.event import Event
from boa3.model.method import Method
from boa3.model.variable import Variable
from boa3.neo.contracts.neffile import NefFile
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest


class TestFileGeneration(BoaTest):
    default_folder: str = 'test_sc/generation_test'

    def test_generate_files(self):
        path = self.get_contract_path('GenerationWithDecorator.py')
        expected_nef_output = path.replace('.py', '.nef')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        Boa3.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_nef_output))
        self.assertTrue(os.path.exists(expected_manifest_output))

    def test_generate_files_with_debug_info(self):
        path = self.get_contract_path('GenerationWithDecorator.py')
        expected_nef_output = path.replace('.py', '.nef')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        expected_debug_info_output = path.replace('.py', '.nefdbgnfo')
        Boa3.compile_and_save(path, debug=True)

        self.assertTrue(os.path.exists(expected_nef_output))
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertTrue(os.path.exists(expected_debug_info_output))

    def test_generate_nef_file(self):
        path = self.get_contract_path('GenerationWithDecorator.py')
        expected_nef_output = path.replace('.py', '.nef')
        Boa3.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_nef_output))
        with open(expected_nef_output, 'rb') as nef_output:
            magic = nef_output.read(constants.SIZE_OF_INT32)
            compiler = nef_output.read(64)
            compiler = compiler.replace(b'\x00', b'')

            nef_output.read(2)  # reserved

            method_token_count = nef_output.read(1)
            self.assertEqual(Integer.from_bytes(method_token_count), 0)

            nef_output.read(2)  # reserved

            script_size = nef_output.read(1)
            script = nef_output.read(Integer.from_bytes(script_size))
            check_sum = Integer.from_bytes(nef_output.read(constants.SIZE_OF_INT32))

        self.assertEqual(Integer.from_bytes(script_size), len(script))

        nef = NefFile(script)._nef
        self.assertEqual(compiler.decode(constants.ENCODING), nef.compiler)
        self.assertEqual(check_sum, nef.checksum)

    def test_generate_manifest_file_with_decorator(self):
        path = self.get_contract_path('GenerationWithDecorator.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entrypoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        # method Main
        method0 = abi['methods'][0]
        self.assertIn('returntype', method0)
        self.assertEqual(AbiType.Integer, method0['returntype'])
        self.assertIn('parameters', method0)
        self.assertEqual(2, len(method0['parameters']))

        arg0 = method0['parameters'][0]
        self.assertIn('name', arg0)
        self.assertEqual('a', arg0['name'])
        self.assertIn('type', arg0)
        self.assertEqual(AbiType.Integer, arg0['type'])

        arg1 = method0['parameters'][1]
        self.assertIn('name', arg1)
        self.assertEqual('b', arg1['name'])
        self.assertIn('type', arg1)
        self.assertEqual(AbiType.Integer, arg1['type'])

        # method Sub
        method1 = abi['methods'][1]
        self.assertIn('returntype', method1)
        self.assertEqual(AbiType.Integer, method1['returntype'])
        self.assertIn('parameters', method1)
        self.assertEqual(2, len(method1['parameters']))

        arg0 = method1['parameters'][0]
        self.assertIn('name', arg0)
        self.assertEqual('a', arg0['name'])
        self.assertIn('type', arg0)
        self.assertEqual(AbiType.Integer, arg0['type'])

        arg1 = method1['parameters'][1]
        self.assertIn('name', arg1)
        self.assertEqual('b', arg1['name'])
        self.assertIn('type', arg1)
        self.assertEqual(AbiType.Integer, arg1['type'])

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_manifest_file_without_decorator(self):
        path = self.get_contract_path('GenerationWithoutDecorator.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(1, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_manifest_file_with_event(self):
        path = self.get_contract_path('test_sc/event_test', 'EventWithArgument.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        events: Dict[str, Event] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Event)
        }

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(1, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        for abi_event in abi['events']:
            self.assertIn('name', abi_event)
            self.assertIn(abi_event['name'], events)
            self.assertIn('parameters', abi_event)

            event_args = events[abi_event['name']].args
            for event_param in abi_event['parameters']:
                self.assertIn('name', event_param)
                self.assertIn(event_param['name'], event_args)
                self.assertIn('type', event_param)
                self.assertEqual(event_args[event_param['name']].type.abi_type,
                                 event_param['type'])

    def test_generate_manifest_file_with_nep5_transfer_event(self):
        path = self.get_contract_path('test_sc/event_test', 'EventNep5Transfer.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        events: Dict[str, Event] = {
            event.name: event
            for event in self.get_compiler_analyser(compiler).symbol_table.values()
            if isinstance(event, Event)
        }

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(1, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        for abi_event in abi['events']:
            self.assertIn('name', abi_event)
            self.assertIn(abi_event['name'], events)
            self.assertIn('parameters', abi_event)

            event_args = events[abi_event['name']].args
            for event_param in abi_event['parameters']:
                self.assertIn('name', event_param)
                self.assertIn(event_param['name'], event_args)
                self.assertIn('type', event_param)
                self.assertEqual(event_args[event_param['name']].type.abi_type,
                                 event_param['type'])

    def test_generate_manifest_file_with_imported_event(self):
        path = self.get_contract_path('GenerationWithImportedEvent.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        events: Dict[str, Event] = {}
        for name, method in self.get_compiler_analyser(compiler).symbol_table.items():
            if isinstance(method, Event):
                events[name] = method
                if method.name not in events:
                    events[method.name] = method

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        for abi_event in abi['events']:
            self.assertIn('name', abi_event)
            self.assertIn(abi_event['name'], events)
            self.assertIn('parameters', abi_event)

            event_args = events[abi_event['name']].args
            for event_param in abi_event['parameters']:
                self.assertIn('name', event_param)
                self.assertIn(event_param['name'], event_args)
                self.assertIn('type', event_param)
                self.assertEqual(event_args[event_param['name']].type.abi_type,
                                 event_param['type'])

    def test_generate_manifest_file_with_public_name_decorator_kwarg(self):
        path = self.get_contract_path('MetadataMethodName.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        # method Main named as Add
        method0 = abi['methods'][0]
        self.assertIn('name', method0)
        self.assertEqual('Add', method0['name'])

    def test_generate_manifest_file_with_public_name_decorator_arg(self):
        path = self.get_contract_path('MetadataMethodNameArg.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        # method Main named as Add
        method0 = abi['methods'][0]
        self.assertIn('name', method0)
        self.assertEqual('Add', method0['name'])

    def test_metadata_abi_method_name_mismatched_type(self):
        path = self.get_contract_path('MetadataMethodNameMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_abi_method_with_duplicated_name_but_different_args(self):
        path = self.get_contract_path('MetadataMethodDuplicatedNameDifferentArgs.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(3, len(abi['methods']))

        # method Inc named as Add
        method0 = abi['methods'][0]
        self.assertIn('name', method0)
        self.assertEqual('Add', method0['name'])
        self.assertIn('parameters', method0)
        self.assertEqual(1, len(method0['parameters']))

        # method Add also named as Add, but with different arg count
        method1 = abi['methods'][1]
        self.assertIn('name', method1)
        self.assertEqual('Add', method1['name'])
        self.assertIn('parameters', method1)
        self.assertEqual(2, len(method1['parameters']))

    def test_metadata_abi_method_with_duplicated_name_and_args(self):
        path = self.get_contract_path('MetadataMethodDuplicatedNameAndArgs.py')
        self.assertCompilerLogs(CompilerError.DuplicatedManifestIdentifier, path)

    def test_generate_manifest_file_with_public_safe_decorator_kwarg(self):
        path = self.get_contract_path('MetadataMethodSafe.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        # method Main
        method0 = abi['methods'][0]
        self.assertIn('safe', method0)
        self.assertEqual(True, method0['safe'])

    def test_generate_manifest_file_with_public_safe_decorator_arg(self):
        path = self.get_contract_path('MetadataMethodSafeArg.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        # method Main named as Add
        method0 = abi['methods'][0]
        self.assertIn('name', method0)
        self.assertEqual('Add', method0['name'])
        self.assertIn('safe', method0)
        self.assertEqual(True, method0['safe'])

    def test_generate_manifest_file_with_unused_event(self):
        path = self.get_contract_path('MetadataUnusedEvent.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(1, len(abi['methods']))

        # notify and uncalled_event shouldn't be added on the abi, because they were not called
        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        event = abi['events'][0]
        self.assertIn('name', event)
        self.assertEqual('CalledEvent', event['name'])
        self.assertIn('parameters', event)
        self.assertEqual(3, len(event['parameters']))
        parameters = event['parameters']
        self.assertEqual(('var4', 'Integer'), (parameters[0]['name'], parameters[0]['type']))
        self.assertEqual(('var5', 'String'), (parameters[1]['name'], parameters[1]['type']))
        self.assertEqual(('var6', 'ByteArray'), (parameters[2]['name'], parameters[2]['type']))

    def test_metadata_abi_method_safe_mismatched_type(self):
        path = self.get_contract_path('MetadataMethodSafeMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_generate_nefdbgnfo_file(self):
        from boa3.model.type.itype import IType
        path = self.get_contract_path('GenerationWithDecorator.py')

        expected_nef_output = path.replace('.py', '.nefdbgnfo')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'), debug=True)
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }

        self.assertTrue(os.path.exists(expected_nef_output))
        debug_info = self.get_debug_info(path)
        self.assertNotIn('entrypoint', debug_info)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        for debug_method in debug_info['methods']:
            self.assertIn('name', debug_method)
            parsed_name = debug_method['name'].split(constants.VARIABLE_NAME_SEPARATOR)
            self.assertEqual(2, len(parsed_name))
            self.assertIn(parsed_name[-1], methods)
            actual_method = methods[parsed_name[-1]]

            # validate id
            self.assertIn('id', debug_method)
            self.assertEqual(str(id(actual_method)), debug_method['id'])

            # validate parameters
            self.assertIn('params', debug_method)
            self.assertEqual(len(actual_method.args), len(debug_method['params']))
            for var in debug_method['params']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                param_id, param_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(param_id, actual_method.args)
                self.assertEqual(param_type, actual_method.args[param_id].type.abi_type)

            # validate local variables
            self.assertIn('variables', debug_method)
            self.assertEqual(len(actual_method.locals), len(debug_method['variables']))
            for var in debug_method['variables']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                var_id, var_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(var_id, actual_method.locals)
                local_type = actual_method.locals[var_id].type
                self.assertEqual(local_type.abi_type if isinstance(local_type, IType) else AbiType.Any, var_type)

    def test_generate_nefdbgnfo_file_with_event(self):
        path = self.get_contract_path('test_sc/event_test', 'EventWithArgument.py')

        expected_nef_output = path.replace('.py', '.nefdbgnfo')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'), debug=True)
        events: Dict[str, Event] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Event)
        }

        self.assertTrue(os.path.exists(expected_nef_output))
        debug_info = self.get_debug_info(path)
        self.assertNotIn('entrypoint', debug_info)
        self.assertIn('events', debug_info)
        self.assertGreater(len(debug_info['events']), 0)

        for debug_event in debug_info['events']:
            self.assertIn('name', debug_event)
            parsed_name = debug_event['name'].split(constants.VARIABLE_NAME_SEPARATOR)
            self.assertEqual(2, len(parsed_name))
            self.assertIn(parsed_name[-1], events)
            actual_event = events[parsed_name[-1]]

            # validate id
            self.assertIn('id', debug_event)
            self.assertEqual(str(id(actual_event)), debug_event['id'])

            # validate parameters
            self.assertIn('params', debug_event)
            self.assertEqual(len(actual_event.args), len(debug_event['params']))
            for var in debug_event['params']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                param_id, param_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(param_id, actual_event.args)
                self.assertEqual(param_type, actual_event.args[param_id].type.abi_type)

    def test_generate_nefdbgnfo_file_with_static_variables(self):
        path = self.get_contract_path('GenerationWithStaticVariables.py')

        expected_nef_output = path.replace('.py', '.nefdbgnfo')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'), debug=True)
        variables: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Variable)
        }

        self.assertTrue(os.path.exists(expected_nef_output))
        debug_info = self.get_debug_info(path)

        self.assertNotIn('entrypoint', debug_info)
        self.assertIn('static-variables', debug_info)
        self.assertGreater(len(debug_info['static-variables']), 0)

        for static_variable in debug_info['static-variables']:
            # validate parameters
            self.assertEqual(3, len(static_variable.split(constants.VARIABLE_NAME_SEPARATOR)))
            var_id, var_type, var_slot = static_variable.split(constants.VARIABLE_NAME_SEPARATOR)
            if var_id not in variables:
                self.assertIn(var_id, [var_full_id.split(constants.VARIABLE_NAME_SEPARATOR)[-1]
                                       for var_full_id in variables])
                var_inner_id = next((var for var in variables
                                     if var.split(constants.VARIABLE_NAME_SEPARATOR)[-1] == var_id),
                                    var_id)
            else:
                var_inner_id = var_id

            self.assertIn(var_inner_id, variables)
            self.assertEqual(var_type, variables[var_inner_id].type.abi_type)

    def test_generate_nefdbgnfo_file_with_user_module_import(self):
        from boa3.model.type.itype import IType
        path = self.get_contract_path('GenerationWithUserModuleImportsDupNames.py')

        expected_nef_output = path.replace('.py', '.nefdbgnfo')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'), debug=True)
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_all_imported_methods(compiler).items()
            if isinstance(method, Method)
        }

        self.assertTrue(os.path.exists(expected_nef_output))
        debug_info = self.get_debug_info(path)
        self.assertNotIn('entrypoint', debug_info)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        for debug_method in debug_info['methods']:
            self.assertIn('name', debug_method)
            name_without_parsing = debug_method['name']
            parsed_name = name_without_parsing.split(constants.VARIABLE_NAME_SEPARATOR)
            self.assertEqual(2, len(parsed_name))
            self.assertIn(name_without_parsing, methods)
            actual_method = methods[name_without_parsing]

            # validate id
            self.assertIn('id', debug_method)
            self.assertEqual(str(id(actual_method)), debug_method['id'])

            # validate parameters
            self.assertIn('params', debug_method)
            self.assertEqual(len(actual_method.args), len(debug_method['params']))
            for var in debug_method['params']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                param_id, param_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(param_id, actual_method.args)
                self.assertEqual(param_type, actual_method.args[param_id].type.abi_type)

            # validate local variables
            self.assertIn('variables', debug_method)
            self.assertEqual(len(actual_method.locals), len(debug_method['variables']))
            for var in debug_method['variables']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                var_id, var_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(var_id, actual_method.locals)
                local_type = actual_method.locals[var_id].type
                self.assertEqual(local_type.abi_type if isinstance(local_type, IType) else AbiType.Any, var_type)

    def test_generate_manifest_file_with_notify_event(self):
        path = self.get_contract_path('test_sc/interop_test/runtime', 'NotifySequence.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('events', abi)
        self.assertGreater(len(abi['events']), 0)

        notify_event = next(abi_event for abi_event in abi['events']
                            if 'name' in abi_event and abi_event['name'] == 'notify')
        self.assertIsNotNone(notify_event, "notify event is not listed in the contract's abi")
        self.assertIn('parameters', notify_event)
        self.assertEqual(1, len(notify_event['parameters']))
        self.assertIn('type', notify_event['parameters'][0])
        self.assertEqual(AbiType.Any, notify_event['parameters'][0]['type'])

    def test_generate_without_main(self):
        path = self.get_contract_path('GenerationWithoutMain.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entryPoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_without_main_and_public_methods(self):
        # since 0.11.2 methods that are not public nor are called are not generated to optimize code
        # this test generates an empty contract, which results in failing compilation
        path = self.get_contract_path('GenerationWithoutMainAndPublicMethods.py')

        with self.assertRaises(NotLoadedException):
            output, manifest = self.compile_and_save(path)

    def test_generate_manifest_file_abi_method_offset(self):
        path = self.get_contract_path('GenerationWithDecorator.py')
        manifest_path = path.replace('.py', '.manifest.json')

        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }
        self.assertGreater(len(methods), 0)

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(manifest_path))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        abi_methods = abi['methods']
        self.assertGreater(len(abi['methods']), 0)

        for method in abi_methods:
            self.assertIn('name', method)
            self.assertIn('offset', method)
            self.assertIn(method['name'], methods)
            self.assertEqual(method['offset'], methods[method['name']].start_address)

    def test_generate_debug_info_with_multiple_flows(self):
        path = self.get_contract_path('GenerationWithMultipleFlows.py')

        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'), debug=True)
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }
        self.assertGreater(len(methods), 0)

        debug_info = self.get_debug_info(path)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        for debug_method in debug_info['methods']:
            self.assertIn('name', debug_method)
            parsed_name = debug_method['name'].split(constants.VARIABLE_NAME_SEPARATOR)
            self.assertEqual(2, len(parsed_name))
            self.assertIn(parsed_name[-1], methods)
            actual_method = methods[parsed_name[-1]]

            # validate id
            self.assertIn('id', debug_method)
            self.assertEqual(str(id(actual_method)), debug_method['id'])

            # validate parameters
            self.assertIn('params', debug_method)
            self.assertEqual(len(actual_method.args), len(debug_method['params']))
            for var in debug_method['params']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                param_id, param_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(param_id, actual_method.args)
                self.assertEqual(param_type, actual_method.args[param_id].type.abi_type)

            # validate local variables
            self.assertIn('variables', debug_method)
            self.assertEqual(len(actual_method.locals), len(debug_method['variables']))
            for var in debug_method['variables']:
                self.assertEqual(2, len(var.split(constants.VARIABLE_NAME_SEPARATOR)))
                var_id, var_type = var.split(constants.VARIABLE_NAME_SEPARATOR)
                self.assertIn(var_id, actual_method.locals)
                self.assertEqual(actual_method.locals[var_id].type.abi_type, var_type)

    def test_generate_init_method(self):
        path = self.get_contract_path('test_sc/variable_test', 'GlobalAssignmentWithType.py')

        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'), debug=True)
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }

        self.assertGreater(len(methods), 0)
        self.assertIn(constants.INITIALIZE_METHOD_ID, methods)
        init_method = methods[constants.INITIALIZE_METHOD_ID]

        output, manifest = self.get_output(path)
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertGreater(len(abi['methods']), 0)

        abi_init = next(method for method in abi['methods']
                        if 'name' in method and method['name'] == constants.INITIALIZE_METHOD_ID)
        self.assertIsNotNone(abi_init)
        self.assertIn('offset', abi_init)
        self.assertEqual(init_method.start_address, abi_init['offset'])
        self.assertIn('parameters', abi_init)
        self.assertEqual(0, len(abi_init['parameters']))
        self.assertIn('returntype', abi_init)
        self.assertEqual(AbiType.Void, abi_init['returntype'])

        debug_info = self.get_debug_info(path)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        debug_method = next((method for method in debug_info['methods']
                             if 'id' in method and method['id'] == str(id(init_method))), None)
        self.assertIsNotNone(debug_method)
        parsed_name = debug_method['name'].split(constants.VARIABLE_NAME_SEPARATOR)
        self.assertEqual(2, len(parsed_name))
        self.assertIn(parsed_name[-1], methods)

        # validate parameters
        self.assertIn('params', debug_method)
        self.assertEqual(0, len(debug_method['params']))

        # validate local variables
        self.assertIn('variables', debug_method)
        self.assertEqual(0, len(debug_method['variables']))

        # validate sequence points
        self.assertIn('sequence-points', debug_method)
        self.assertEqual(0, len(debug_method['sequence-points']))

    def test_generate_with_user_module_import(self):
        path = self.get_contract_path('GenerationWithUserModuleImports.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entryPoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

    def test_generate_with_user_module_import_with_project_root(self):
        path = self.get_contract_path('project_path', 'GenerationWithUserModuleImportsFromProjectRoot.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path, root_folder=self.get_dir_path(self.default_test_folder))

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entryPoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

    def test_compiler_error(self):
        path = self.get_contract_path('test_sc/built_in_methods_test', 'ClearTooManyParameters.py')

        with self.assertRaises(NotLoadedException):
            Boa3.compile(path)

        with self.assertRaises(NotLoadedException):
            Boa3.compile_and_save(path)

        with self.assertRaises(NotLoadedException):
            self.compile_and_save(path)

    def test_generation_with_recursive_function(self):
        path = self.get_contract_path('test_sc/function_test', 'RecursiveFunction.py')

        from boa3_test.tests.test_classes.testengine import TestEngine
        engine = TestEngine()

        expected = self.fact(57)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected, result)

    def fact(self, f: int) -> int:
        if f <= 1:
            return 1
        return f * self.fact(f - 1)



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.iterator.iteratortype import IteratorType
from boa3.model.variable import Variable


class IteratorNextMethod(InteropMethod):
    def __init__(self):
        from boa3.model.type.type import Type
        syscall = 'System.Iterator.Next'
        identifier = '-iterator_next'
        args: Dict[str, Variable] = {'self': Variable(IteratorType.build())}
        super().__init__(identifier, syscall, args, return_type=Type.bool)



from typing import List

from boa3 import constants
from boa3.boa3 import Boa3
from boa3.neo import to_script_hash
from boa3.neo.cryptography import hash160
from boa3.neo.smart_contract.notification import Notification
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestHTLCTemplate(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')
    OTHER_ACCOUNT_2 = bytes(range(20))

    def test_htlc_compile(self):
        path = self.get_contract_path('htlc.py')
        Boa3.compile(path)

    def test_htlc_atomic_swap(self):
        path = self.get_contract_path('htlc.py')
        engine = TestEngine()

        # starting atomic swap by using the atomic_swap method
        result = self.run_smart_contract(engine, path, 'atomic_swap',
                                         self.OWNER_SCRIPT_HASH, constants.NEO_SCRIPT, 10 * 10 ** 8,
                                         self.OTHER_ACCOUNT_1, constants.GAS_SCRIPT, 10000 * 10 ** 8,
                                         hash160(String('unit test').to_bytes()),
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

    def test_htlc_on_nep17_payment(self):
        path = self.get_contract_path('htlc.py')
        engine = TestEngine()
        transferred_amount_neo = 10 * 10 ** 8
        transferred_amount_gas = 10000 * 10 ** 8

        self.run_smart_contract(engine, path, 'refund')
        htlc_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_neo(self.OTHER_ACCOUNT_1, transferred_amount_neo)
        engine.add_gas(self.OTHER_ACCOUNT_2, transferred_amount_gas)

        # starting atomic swap
        result = self.run_smart_contract(engine, path, 'atomic_swap',
                                         self.OTHER_ACCOUNT_1, constants.NEO_SCRIPT, transferred_amount_neo,
                                         self.OTHER_ACCOUNT_2, constants.GAS_SCRIPT, transferred_amount_gas,
                                         hash160(String('unit test').to_bytes()),
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # transfer wil be aborted at onPayment if the transfer is not valid
        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                    self.OTHER_ACCOUNT_1, htlc_address,
                                    transferred_amount_neo - 100, None,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        # since the transfer was aborted, it was not registered in the events
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(0, len(transfer_events))

        # saving the balance to compare after the transfer
        balance_neo_sender_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                            self.OTHER_ACCOUNT_1)
        balance_neo_receiver_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', htlc_address)

        # this transfer will be accepted
        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, htlc_address, transferred_amount_neo, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # transfer was accepted so it was registered
        transfer_events = engine.get_events('Transfer', constants.NEO_SCRIPT)
        self.assertEqual(1, len(transfer_events))
        neo_transfer_event = transfer_events[0]
        self.assertEqual(3, len(neo_transfer_event.arguments))

        sender, receiver, amount = neo_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, sender)
        self.assertEqual(htlc_address, receiver)
        self.assertEqual(transferred_amount_neo, amount)

        # saving the balance after to compare with the balance before the transfer
        balance_neo_sender_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                           self.OTHER_ACCOUNT_1)
        balance_neo_receiver_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', htlc_address)

        self.assertEqual(balance_neo_sender_before - transferred_amount_neo, balance_neo_sender_after)
        self.assertEqual(balance_neo_receiver_before + transferred_amount_neo, balance_neo_receiver_after)

        # transfer won't be accepted, because amount is wrong
        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                    self.OTHER_ACCOUNT_2, htlc_address,
                                    transferred_amount_gas - 100, None,
                                    signer_accounts=[self.OTHER_ACCOUNT_2])

        transfer_events = engine.get_events('Transfer', origin=constants.NEO_SCRIPT)
        # the NEO transfer
        self.assertEqual(1, len(transfer_events))

        # saving the balance to compare after the transfer
        balance_gas_sender_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                            self.OTHER_ACCOUNT_2)
        balance_gas_receiver_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf', htlc_address)

        # this transfer will be accepted
        result = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_2, htlc_address, transferred_amount_gas, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_2])
        self.assertEqual(True, result)

        # the transfer was accepted so it was registered
        transfer_events = engine.get_events('Transfer', origin=constants.GAS_SCRIPT)
        self.assertGreaterEqual(len(transfer_events), 1)
        gas_transfer_event = transfer_events[-1]
        self.assertEqual(3, len(gas_transfer_event.arguments))

        sender, receiver, amount = gas_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_2, sender)
        self.assertEqual(htlc_address, receiver)
        self.assertEqual(transferred_amount_gas, amount)

        # saving the balance after to compare with the balance before the transfer
        balance_gas_sender_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                           self.OTHER_ACCOUNT_2)
        balance_gas_receiver_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf', htlc_address)

        self.assertEqual(balance_gas_sender_before - transferred_amount_gas, balance_gas_sender_after)
        self.assertEqual(balance_gas_receiver_before + transferred_amount_gas, balance_gas_receiver_after)

    def test_htlc_withdraw(self):
        path = self.get_contract_path('htlc.py')
        engine = TestEngine()
        transferred_amount_neo = 10 * 10 ** 8
        transferred_amount_gas = 10000 * 10 ** 8

        self.run_smart_contract(engine, path, 'refund')
        htlc_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_neo(self.OTHER_ACCOUNT_1, transferred_amount_neo)
        engine.add_gas(self.OTHER_ACCOUNT_2, transferred_amount_gas)

        # saving the balance to compare after the withdraw
        balance_neo_person_a_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                              self.OTHER_ACCOUNT_1)
        balance_neo_person_b_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                              self.OTHER_ACCOUNT_2)
        balance_neo_htlc_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', htlc_address)
        balance_gas_person_a_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                              self.OTHER_ACCOUNT_1)
        balance_gas_person_b_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                              self.OTHER_ACCOUNT_2)
        balance_gas_htlc_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf', htlc_address)

        # starting atomic swap by using the atomic_swap method
        result = self.run_smart_contract(engine, path, 'atomic_swap',
                                         self.OTHER_ACCOUNT_1, constants.NEO_SCRIPT, transferred_amount_neo,
                                         self.OTHER_ACCOUNT_2, constants.GAS_SCRIPT, transferred_amount_gas,
                                         hash160(String('unit test').to_bytes()),
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # won't be able to withdraw, because no one transferred cryptocurrency to the smart contract
        result = self.run_smart_contract(engine, path, 'withdraw', 'unit test',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, htlc_address, transferred_amount_neo, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer', origin=constants.NEO_SCRIPT)
        self.assertEqual(1, len(transfer_events))

        result = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_2, htlc_address, transferred_amount_gas, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_2])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer', constants.GAS_SCRIPT)
        self.assertEqual(2, len(transfer_events))

        # the withdraw will fail, because the secret is wrong
        result = self.run_smart_contract(engine, path, 'withdraw', 'wrong one',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        # the withdraw will occur
        result = self.run_smart_contract(engine, path, 'withdraw', 'unit test',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # the transfer were accepted so they were registered
        transfer_events = engine.get_events('Transfer', origin=constants.GAS_SCRIPT)
        self.assertGreaterEqual(len(transfer_events), 1)
        gas_transfer_event = transfer_events[-1]
        self.assertEqual(3, len(gas_transfer_event.arguments))

        transfer_events = engine.get_events('Transfer', constants.NEO_SCRIPT)
        self.assertGreaterEqual(len(transfer_events), 1)
        neo_transfer_event = transfer_events[-1]
        self.assertEqual(3, len(neo_transfer_event.arguments))

        sender, receiver, amount = gas_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(htlc_address, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(transferred_amount_gas, amount)

        sender, receiver, amount = neo_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(htlc_address, sender)
        self.assertEqual(self.OTHER_ACCOUNT_2, receiver)
        self.assertEqual(transferred_amount_neo, amount)

        # saving the balance after to compare with the balance before the transfer
        minted_gas_per_neo_transfer = int(transferred_amount_neo * 0.5)
        balance_neo_person_a_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                             self.OTHER_ACCOUNT_1)
        balance_neo_person_b_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                             self.OTHER_ACCOUNT_2)
        balance_neo_htlc_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', htlc_address)

        balance_gas_person_a_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                             self.OTHER_ACCOUNT_1)
        balance_gas_person_b_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                             self.OTHER_ACCOUNT_2)
        balance_gas_htlc_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf', htlc_address)

        self.assertEqual(balance_neo_person_a_before - transferred_amount_neo, balance_neo_person_a_after)
        self.assertEqual(balance_neo_person_b_before + transferred_amount_neo, balance_neo_person_b_after)
        self.assertEqual(balance_neo_htlc_before, balance_neo_htlc_after)
        self.assertEqual(balance_gas_person_a_before + transferred_amount_gas + minted_gas_per_neo_transfer,
                         balance_gas_person_a_after)
        self.assertEqual(balance_gas_person_b_before - transferred_amount_gas, balance_gas_person_b_after)
        self.assertEqual(balance_gas_htlc_before, balance_gas_htlc_after)

    def test_htlc_refund(self):
        path = self.get_contract_path('htlc.py')
        engine = TestEngine()
        transferred_amount_neo = 10 * 10 ** 8
        transferred_amount_gas = 10000 * 10 ** 8

        self.run_smart_contract(engine, path, 'refund')
        htlc_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_neo(self.OTHER_ACCOUNT_1, transferred_amount_neo)
        engine.add_gas(self.OTHER_ACCOUNT_2, transferred_amount_gas)

        result = self.run_smart_contract(engine, path, 'atomic_swap',
                                         self.OTHER_ACCOUNT_1, constants.NEO_SCRIPT, transferred_amount_neo,
                                         self.OTHER_ACCOUNT_2, constants.GAS_SCRIPT, transferred_amount_gas,
                                         hash160(String('unit test').to_bytes()),
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # won't be able to refund, because not enough time has passed
        result = self.run_smart_contract(engine, path, 'refund',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(False, result)

        # this simulates a new block in the blockchain
        # time only changes value when a new block enters the blockchain
        engine.increase_block()
        # will be able to refund, because enough time has passed
        result = self.run_smart_contract(engine, path, 'refund',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        transfer_events: List[Notification] = []
        # no one transferred cryptocurrency to the contract, so no one was refunded and no Transfer occurred
        # removing possible GAS minting from the List
        for k in engine.get_events('Transfer'):
            if k.arguments[0] is not None:
                transfer_events.append(k)
        self.assertEqual(0, len(transfer_events))

        # starting atomic swap by using the atomic_swap method
        result = self.run_smart_contract(engine, path, 'atomic_swap',
                                         self.OTHER_ACCOUNT_1, constants.NEO_SCRIPT, transferred_amount_neo,
                                         self.OTHER_ACCOUNT_2, constants.GAS_SCRIPT, transferred_amount_gas,
                                         hash160(String('unit test').to_bytes()),
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, htlc_address, transferred_amount_neo, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        for k in engine.get_events('Transfer'):
            if k.arguments[0] is not None:
                transfer_events.append(k)
        self.assertEqual(1, len(transfer_events))
        self.assertEqual(3, len(transfer_events[0].arguments))

        engine.increase_block()
        # will be able to refund, because enough time has passed
        result = self.run_smart_contract(engine, path, 'refund',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # OWNER transferred cryptocurrency to the contract, so only he will be refunded
        transfer_events = []
        # removing possible GAS minting from the List
        for k in engine.get_events('Transfer'):
            if k.arguments[0] is not None:
                transfer_events.append(k)
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        # HTLC returning the tokens
        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(htlc_address, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(transferred_amount_neo, amount)

        result = self.run_smart_contract(engine, path, 'atomic_swap',
                                         self.OTHER_ACCOUNT_1, constants.NEO_SCRIPT, transferred_amount_neo,
                                         self.OTHER_ACCOUNT_2, constants.GAS_SCRIPT, transferred_amount_gas,
                                         hash160(String('unit test').to_bytes()),
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, htlc_address, transferred_amount_neo, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        transfer_events = []
        for k in engine.get_events('Transfer'):
            if k.arguments[0] is not None:
                transfer_events.append(k)
        self.assertEqual(3, len(transfer_events))
        self.assertEqual(3, len(transfer_events[2].arguments))

        result = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_2, htlc_address, transferred_amount_gas, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_2])
        self.assertEqual(True, result)
        transfer_events = []
        for k in engine.get_events('Transfer'):
            if k.arguments[0] is not None:
                transfer_events.append(k)
        self.assertEqual(4, len(transfer_events))
        self.assertEqual(3, len(transfer_events[3].arguments))

        engine.increase_block()
        # will be able to refund, because enough time has passed
        result = self.run_smart_contract(engine, path, 'refund',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # OWNER and OTHER_ACCOUNT transferred cryptocurrency to the contract, so they both will be refunded
        transfer_events = []
        # removing possible GAS minting from the List
        for k in engine.get_events('Transfer'):
            if k.arguments[0] is not None:
                transfer_events.append(k)
        self.assertEqual(6, len(transfer_events))
        self.assertEqual(3, len(transfer_events[4].arguments))
        self.assertEqual(3, len(transfer_events[5].arguments))

        sender, receiver, amount = transfer_events[4].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(htlc_address, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(transferred_amount_neo, amount)

        sender, receiver, amount = transfer_events[5].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(htlc_address, sender)
        self.assertEqual(self.OTHER_ACCOUNT_2, receiver)
        self.assertEqual(transferred_amount_gas, amount)



from __future__ import annotations

from decimal import Decimal, localcontext
from typing import Type


class BigInteger(int):
    """
    An implementation to match the C# BigInteger class as used by the NEO reference project.
    """

    @classmethod
    def ZERO(cls: Type[BigInteger]) -> BigInteger:
        """
        Returns:
            An instance initialized to zero.
        """
        return cls(0)

    @classmethod
    def ONE(cls: Type[BigInteger]) -> BigInteger:
        """
        Returns:
            An instance initialized to one.
        """
        return cls(1)

    @classmethod
    def frombytes(cls: Type[BigInteger], bytes: bytes) -> BigInteger:
        """
        Return the BigInteger represented by the given array of bytes.

        The underlying value is always signed and in little endian format.

        Args:
            bytes: Holds the array of bytes to convert. The argument must either support the buffer protocol or
               be an iterable object producing bytes.
        """
        return cls(int.from_bytes(bytes, 'little', signed=True))

    @property
    def sign(self) -> int:
        """
        Gets a number that indicates the sign (negative, positive, or zero) of the current BigInteger object.

        Returns:
            - -1 The value of this object is negative.
            - 0 The value of this object is 0 (zero).
            - 1  The value of this object is positive.
        """
        if self > 0:
            return 1
        elif self == 0:
            return 0
        return -1

    def to_bytearray(self) -> bytearray:
        """
        Return an array of bytes representing the BigInteger.
        """
        if self == 0:
            return bytearray(b'\x00')

        if self < 0:
            highbyte = 0xff
            data = bytearray(self.to_bytes(1 + ((self.bit_length() + 7) // 8), byteorder='little', signed=True))

            msb = len(data) - 1
            for i, b in enumerate(data[::-1]):
                if b != highbyte:
                    msb -= i
                    break

            needExtraByte = (data[msb] & 0x80) != (highbyte & 0x80)
            if needExtraByte:
                return data
            else:
                return data[:-1]

        try:
            return bytearray(self.to_bytes((self.bit_length() + 7) // 8, byteorder='little', signed=True))
        except OverflowError:
            return bytearray(self.to_bytes(1 + ((self.bit_length() + 7) // 8), byteorder='little', signed=True))

    def __abs__(self, *args, **kwargs):
        return BigInteger(super(BigInteger, self).__abs__(*args, **kwargs))

    def __add__(self, *args, **kwargs):
        return BigInteger(super(BigInteger, self).__add__(*args, **kwargs))

    def __mod__(self, *args, **kwargs):
        with localcontext() as ctx:
            ctx.prec = 100
            d1 = Decimal(self)
            d2 = Decimal(args[0])
            res = int(d1 % d2)
        return BigInteger(res)

    def __mul__(self, *args, **kwargs):
        return BigInteger(super(BigInteger, self).__mul__(*args, **kwargs))

    def __neg__(self, *args, **kwargs):
        return BigInteger(super(BigInteger, self).__neg__(*args, **kwargs))

    def __str__(self, *args, **kwargs):
        return super(BigInteger, self).__str__(*args, **kwargs)

    def __sub__(self, *args, **kwargs):
        return BigInteger(super(BigInteger, self).__sub__(*args, **kwargs))

    def __floordiv__(self, *args, **kwargs):
        return BigInteger(super(BigInteger, self).__floordiv__(*args, **kwargs))

    def __truediv__(self, *args, **kwargs):
        if self < 0 or args[0] < 0:
            return BigInteger(super(BigInteger, self).__truediv__(*args, **kwargs))
        else:
            return BigInteger(super(BigInteger, self).__floordiv__(*args, **kwargs))

    def __rshift__(self, *args, **kwargs):
        shift = args[0]
        if shift < 0:
            shift = abs(shift)
            return BigInteger(super(BigInteger, self).__lshift__(shift, **kwargs))
        else:
            return BigInteger(super(BigInteger, self).__rshift__(*args, **kwargs))

    def __lshift__(self, *args, **kwargs):
        shift = args[0]
        if shift < 0:
            shift = abs(shift)
            return BigInteger(super(BigInteger, self).__rshift__(shift, **kwargs))
        else:
            return BigInteger(super(BigInteger, self).__lshift__(*args, **kwargs))



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestNone(BoaTest):
    default_folder: str = 'test_sc/none_test'

    def test_variable_none(self):
        path = self.get_contract_path('VariableNone.py')

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHNULL
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_none_tuple(self):
        path = self.get_contract_path('NoneTuple.py')

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHNULL   # a = (None, None, None)
            + Opcode.PUSHNULL
            + Opcode.PUSHNULL
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_none_identity(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.ISNULL
            + Opcode.RET        # return
        )

        path = self.get_contract_path('NoneIdentity.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', None)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', '5')
        self.assertEqual(False, result)

    def test_none_not_identity(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.ISNULL
            + Opcode.NOT
            + Opcode.RET        # return
        )

        path = self.get_contract_path('NoneNotIdentity.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', None)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', '5')
        self.assertEqual(True, result)

    def test_none_equality(self):
        path = self.get_contract_path('NoneEquality.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_mismatched_type_int_operation(self):
        path = self.get_contract_path('MismatchedTypesInOperation.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_reassign_variable_with_none(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH2          # a = 2
            + Opcode.STLOC0
            + Opcode.PUSH4          # b = a * 2
            + Opcode.STLOC1
            + Opcode.PUSHNULL       # a = None
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('ReassignVariableWithNone.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_reassign_variable_after_none(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHNULL       # a = None
            + Opcode.STLOC0
            + Opcode.PUSH2          # a = 2
            + Opcode.STLOC0
            + Opcode.PUSH4          # b = a * 2
            + Opcode.STLOC1
            + Opcode.RET        # return
        )
        path = self.get_contract_path('ReassignVariableAfterNone.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_boa2_none_test(self):
        path = self.get_contract_path('NoneBoa2Test.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)



from typing import Dict

from boa3.model.builtin.interop.nativecontract import PolicyContractMethod
from boa3.model.variable import Variable


class IsBlockedMethod(PolicyContractMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'is_blocked'
        native_identifier = 'isBlocked'
        args: Dict[str, Variable] = {
            'account': Variable(UInt160Type.build())
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.bool)



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.type.classes.classarraytype import ClassArrayType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class WitnessRuleType(ClassArrayType):
    """
    A class used to represent Neo WitnessRule class
    """

    def __init__(self):
        super().__init__('WitnessRule')
        from boa3.model.builtin.interop.blockchain.witnessruleactiontype import WitnessRuleActionType
        from boa3.model.builtin.interop.blockchain.witnessconditiontype import WitnessConditionType

        self._variables: Dict[str, Variable] = {
            'action': Variable(WitnessRuleActionType.build()),
            'condition': Variable(WitnessConditionType.build())
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        # was having a problem with recursive import
        if self._constructor is None:
            self._constructor: Method = WitnessRuleMethod(self)
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> WitnessRuleType:
        if value is None or cls._is_type_of(value):
            return _WitnessRule

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, WitnessRuleType)


_WitnessRule = WitnessRuleType()


class WitnessRuleMethod(IBuiltinMethod):

    def __init__(self, return_type: WitnessRuleType):
        identifier = '-WitnessRule__init__'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 0

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.builtin.interop.blockchain.witnessconditiontype import WitnessConditionType
        from boa3.model.builtin.interop.blockchain.witnessruleactiontype import WitnessRuleActionType

        return (WitnessConditionType.build().constructor_method().opcode +  # condition
                WitnessRuleActionType.build().constructor_method().opcode +  # action
                [(Opcode.PUSH1, b''),
                 (Opcode.PACK, b'')
                 ]
                )

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



from typing import Any, Dict, List, Optional, Sized, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.icollection import ICollectionType
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class LenMethod(IBuiltinMethod):

    def __init__(self, collection_type: IType = None):
        from boa3.model.type.type import Type
        if not isinstance(collection_type, ICollectionType):
            collection_type = Type.sequence

        identifier = 'len'
        args: Dict[str, Variable] = {'__o': Variable(collection_type)}
        super().__init__(identifier, args, return_type=Type.int)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.SIZE, b'')]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['__o'].type):
            return self

        if isinstance(value, Sized) and len(value) == 1:
            value = value[0]
        if isinstance(value, ICollectionType):
            return LenMethod(value)
        return super().build(value)



from typing import Any, List

from boa3.builtin.compile_time import \
    public as public_method  # alias to not change other tests when executing lint process


@public_method
def EmptyList() -> List[Any]:
    return []


empty_list = []



from typing import Optional

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.nativecontract.Nep17.getnep17scripthashmethod import GetNep17ScriptHashMethod


class GetNeoScriptHashMethod(GetNep17ScriptHashMethod):
    def __init__(self):
        from boa3.constants import NEO_SCRIPT
        identifier = '-get_neo'
        super().__init__(NEO_SCRIPT, identifier)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None


class NeoProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'NEO'
        getter = GetNeoScriptHashMethod()
        super().__init__(identifier, getter)


NeoToken = NeoProperty()



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestFunction(BoaTest):
    default_folder: str = 'test_sc/function_test'

    def test_integer_function(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x00'  # num local variables
            + b'\x01'  # num arguments
            + Opcode.PUSH10  # body
            + Opcode.RET  # return
        )

        path = self.get_contract_path('IntegerFunction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', -1)
        self.assertEqual(10, result)

    def test_string_function(self):
        expected_output = (
            # functions without arguments and local variables don't need initslot
            Opcode.PUSHDATA1  # body
            + bytes([len('42')])
            + bytes('42', constants.ENCODING)
            + Opcode.RET  # return
        )

        path = self.get_contract_path('StringFunction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('42', result)

    def test_bool_function(self):
        expected_output = (
            # functions without arguments and local variables don't need initslot
            Opcode.PUSHT  # body
            + Opcode.RET  # return
        )

        path = self.get_contract_path('BoolFunction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_none_function_pass(self):
        path = self.get_contract_path('NoneFunctionPass.py')
        engine = TestEngine()

        output = Boa3.compile(path)
        self.assertIn(Opcode.NOP, output)

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertIsVoid(result)

    def test_none_function_return_none(self):
        path = self.get_contract_path('NoneFunctionReturnNone.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsVoid(result)

    def test_none_function_changing_values_with_return(self):
        path = self.get_contract_path('NoneFunctionChangingValuesWithReturn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual([2, 4, 6, 8, 10], result)

    def test_none_function_changing_values_without_return(self):
        path = self.get_contract_path('NoneFunctionChangingValuesWithoutReturn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual([2, 4, 6, 8, 10], result)

    def test_arg_without_type_hint(self):
        path = self.get_contract_path('ArgWithoutTypeHintFunction.py')
        self.assertCompilerLogs(CompilerError.TypeHintMissing, path)

    def test_no_return_hint_function_with_empty_return_statement(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x00'  # num local variables
            + b'\x01'  # num arguments
            + Opcode.RET  # return
        )

        path = self.get_contract_path('EmptyReturnFunction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertIsVoid(result)

    def test_no_return_hint_function_with_condition_empty_return_statement(self):
        path = self.get_contract_path('ConditionEmptyReturnFunction.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'Main', 50)
        self.assertIsVoid(result)

    def test_empty_return_with_optional_return_type(self):
        path = self.get_contract_path('EmptyReturnWithOptionalReturnType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertIsNone(result)

        result = self.run_smart_contract(engine, path, 'Main', 50)
        self.assertEqual(25, result)

    def test_no_return_hint_function_without_return_statement(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x00'  # num local variables
            + b'\x01'  # num arguments
            + Opcode.NOP  # pass
            + Opcode.RET  # return
        )

        path = self.get_contract_path('NoReturnFunction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertIsVoid(result)

    def test_return_type_hint_function_with_empty_return(self):
        path = self.get_contract_path('ExpectingReturnFunction.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_multiple_return_function(self):
        path = self.get_contract_path('MultipleReturnFunction.py')
        self.assertCompilerLogs(CompilerError.TooManyReturns, path)

    def test_tuple_function(self):
        path = self.get_contract_path('TupleFunction.py')
        self.assertCompilerLogs(CompilerError.TooManyReturns, path)

    def test_default_return(self):
        path = self.get_contract_path('DefaultReturn.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_empty_list_return(self):
        expected_output = (
            Opcode.NEWARRAY0
            + Opcode.RET
        )

        path = self.get_contract_path('EmptyListReturn.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_mismatched_return_type(self):
        path = self.get_contract_path('MismatchedReturnType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_mismatched_return_type_with_if(self):
        path = self.get_contract_path('MismatchedReturnTypeWithIf.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_call_void_function_without_args(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.CALL  # TestFunction()
            + called_function_address
            + Opcode.PUSHT  # return True
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1  # a = 1
            + Opcode.STLOC0
            + Opcode.RET  # return
        )

        path = self.get_contract_path('CallVoidFunctionWithoutArgs.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(1, result)

    def test_call_function_without_args(self):
        called_function_address = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x01'
            + b'\x00'
            + Opcode.CALL  # a = TestFunction()
            + called_function_address
            + Opcode.STLOC0
            + Opcode.LDLOC0  # return a
            + Opcode.RET
            + Opcode.PUSH1  # TestFunction
            + Opcode.RET  # return 1
        )

        path = self.get_contract_path('CallReturnFunctionWithoutArgs.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(1, result)

    def test_call_void_function_with_literal_args(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.PUSH2  # TestAdd(1, 2)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.PUSHT  # return True
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x01'
            + b'\x02'
            + Opcode.LDARG0  # c = a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.RET  # return
        )

        path = self.get_contract_path('CallVoidFunctionWithLiteralArgs.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(1, result)

    def test_call_function_with_literal_args(self):
        called_function_address = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH2  # a = TestAdd(1, 2)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.STLOC0
            + Opcode.LDLOC0  # return a
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0  # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET  # return
        )

        path = self.get_contract_path('CallReturnFunctionWithLiteralArgs.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(3, result)

    def test_call_void_function_with_variable_args(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH1  # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2  # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH2  # TestAdd(a, b)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.PUSHT  # return True
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x01'
            + b'\x02'
            + Opcode.LDARG0  # c = a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.RET  # return
        )

        path = self.get_contract_path('CallVoidFunctionWithVariableArgs.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(1, result)

    def test_call_function_with_variable_args(self):
        called_function_address = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x03'
            + b'\x02'
            + Opcode.PUSH1  # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2  # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH2  # c = TestAdd(a, b)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.STLOC2
            + Opcode.LDLOC2  # return c
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0  # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET  # return
        )

        path = self.get_contract_path('CallReturnFunctionWithVariableArgs.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'TestAdd', 1, 2)
        self.assertEqual(1, len(engine.result_stack))
        self.assertEqual(3, engine.result_stack[-1])

    def test_call_function_on_return(self):
        called_function_address = Integer(3).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH1  # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2  # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH2  # return TestAdd(a, b)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0  # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET  # return
        )

        path = self.get_contract_path('CallReturnFunctionOnReturn.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(3, result)

    def test_call_function_without_variables(self):
        main_to_one_address = Integer(-10).to_byte_array(min_length=1, signed=True)
        main_to_two_address = Integer(5).to_byte_array(min_length=1, signed=True)
        two_to_one_address = Integer(-24).to_byte_array(min_length=1, signed=True)
        end_if = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.PUSH1  # One
            + Opcode.RET  # return 1
            + Opcode.INITSLOT  # Main
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0  # if arg0 == 1
            + Opcode.PUSH1
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT
            + end_if
            + Opcode.CALL  # return One()
            + main_to_one_address
            + Opcode.RET
            + Opcode.LDARG0  # elif arg0 == 2
            + Opcode.PUSH2
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT
            + end_if
            + Opcode.CALL  # return Two()
            + main_to_two_address
            + Opcode.RET
            + Opcode.PUSH0  # default return
            + Opcode.RET
            + Opcode.PUSH1  # Two
            + Opcode.CALL  # return 1 + One()
            + two_to_one_address
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('CallFunctionWithoutVariables.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 1)
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', 2)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', 3)
        self.assertEqual(0, result)

    def test_call_function_written_before_caller(self):
        call_address = Integer(-9).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT  # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0  # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET
            + Opcode.PUSH2  # return TestAdd(a, b)
            + Opcode.PUSH1
            + Opcode.CALL
            + call_address
            + Opcode.RET
        )

        path = self.get_contract_path('CallFunctionWrittenBefore.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(3, result)

    def test_return_void_function(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.CALL  # Main
            + called_function_address  # return TestFunction()
            + Opcode.PUSHNULL
            + Opcode.RET
            + Opcode.INITSLOT  # TestFunction
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1  # a = 1
            + Opcode.STLOC0
            + Opcode.RET  # return
        )

        path = self.get_contract_path('ReturnVoidFunction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsNone(result)

    def test_return_void_function_mismatched_type(self):
        path = self.get_contract_path('ReturnVoidFunctionMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_return_inside_if(self):
        path = self.get_contract_path('ReturnIf.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 4)
        self.assertEqual(3, result)
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(6, result)
        result = self.run_smart_contract(engine, path, 'Main', 6)
        self.assertEqual(6, result)

    def test_missing_return_inside_if(self):
        path = self.get_contract_path('ReturnIfMissing.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_missing_return_inside_elif(self):
        path = self.get_contract_path('ReturnElifMissing.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_missing_return_inside_else(self):
        path = self.get_contract_path('ReturnElseMissing.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_return_inside_multiple_inner_if(self):
        path = self.get_contract_path('ReturnMultipleInnerIf.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(9, result)

    def test_missing_return_inside_multiple_inner_if(self):
        path = self.get_contract_path('ReturnMultipleInnerIfMissing.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_return_if_expression(self):
        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0  # return 5 if condition else 10
            + Opcode.JMPIFNOT
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH5  # 5
            + Opcode.JMP  # else
            + Integer(3).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH10  # 10
            + Opcode.RET  # return
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnIfExpression.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(5, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(10, result)

    def test_return_if_expression_mismatched_type(self):
        path = self.get_contract_path('ReturnIfExpressionMismatched.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_return_inside_for(self):
        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0  # for_sequence = arg0
            + Opcode.PUSH0  # for_index = 0
            + Opcode.JMP  # begin for
            + Integer(19).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER  # value = for_sequence[for_index]
            + Opcode.OVER
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC0
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0  # return value
            + Opcode.RET
            + Opcode.INC  # for_index = for_index + 1
            + Opcode.DUP  # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-22).to_byte_array(min_length=1, signed=True)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.PUSH5  # else
            + Opcode.RET  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnFor.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', [1, 2, 3])
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', (2, 5, 7))
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', [])
        self.assertEqual(5, result)

    def test_missing_return_inside_for(self):
        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x02'
            + b'\x01'
            + Opcode.PUSH0  # x = 0
            + Opcode.STLOC0
            + Opcode.LDARG0  # for_sequence = arg0
            + Opcode.PUSH0  # for_index = 0
            + Opcode.JMP  # begin for
            + Integer(19).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER  # value = for_sequence[for_index]
            + Opcode.OVER
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC1
            + Opcode.LDLOC0  # x += value
            + Opcode.LDLOC1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.INC  # for_index = for_index + 1
            + Opcode.DUP  # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-22).to_byte_array(min_length=1, signed=True)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0  # else
            + Opcode.RET  # return x
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnForOnlyOnElse.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', [1, 2, 3])
        self.assertEqual(6, result)
        result = self.run_smart_contract(engine, path, 'Main', (2, 5, 7))
        self.assertEqual(14, result)
        result = self.run_smart_contract(engine, path, 'Main', [])
        self.assertEqual(0, result)

    def test_missing_return_inside_for_else(self):
        path = self.get_contract_path('ReturnForElseMissing.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_return_inside_while(self):
        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0  # x = arg0
            + Opcode.STLOC0
            + Opcode.JMP  # begin while
            + Integer(8).to_byte_array(min_length=1, signed=True)
            + Opcode.LDLOC0  # x += 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0  # return x
            + Opcode.RET
            + Opcode.LDLOC0
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIF  # end while x < 10
            + Integer(-9).to_byte_array(min_length=1, signed=True)
            + Opcode.LDLOC0  # else
            + Opcode.RET  # return x
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnWhile.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(6, result)
        result = self.run_smart_contract(engine, path, 'Main', 3)
        self.assertEqual(4, result)
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', 100)
        self.assertEqual(100, result)

    def test_missing_return_inside_while(self):
        expected_output = (
            Opcode.INITSLOT  # Main
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0  # x = arg0
            + Opcode.STLOC0
            + Opcode.JMP  # begin while
            + Integer(6).to_byte_array(min_length=1, signed=True)
            + Opcode.LDLOC0  # x += 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIF  # end while x < 10
            + Integer(-7).to_byte_array(min_length=1, signed=True)
            + Opcode.LDLOC0  # else
            + Opcode.RET  # return x
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnWhileOnlyOnElse.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', 3)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', 100)
        self.assertEqual(100, result)

    def test_missing_return_inside_while_without_else(self):
        path = self.get_contract_path('ReturnWhileWithoutElse.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_multiple_function_large_call(self):
        path = self.get_contract_path('MultipleFunctionLargeCall.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'calculate', [1])
        self.assertIsNone(result)

        result = self.run_smart_contract(engine, path, 'main', 'calc', [1, 2, 3])
        self.assertIsNone(result)

        result = self.run_smart_contract(engine, path, 'calculate', 1, [10, 3])
        self.assertEqual(13, result)
        result = self.run_smart_contract(engine, path, 'main', 'calculate', [1, 10, 3])
        self.assertEqual(13, result)

        result = self.run_smart_contract(engine, path, 'calculate', 2, [10, 3])
        self.assertEqual(7, result)
        result = self.run_smart_contract(engine, path, 'main', 'calculate', [2, 10, 3])
        self.assertEqual(7, result)

        result = self.run_smart_contract(engine, path, 'calculate', 3, [10, 3])
        self.assertEqual(3, result)
        result = self.run_smart_contract(engine, path, 'main', 'calculate', [3, 10, 3])
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'calculate', 4, [10, 3])
        self.assertEqual(30, result)
        result = self.run_smart_contract(engine, path, 'main', 'calculate', [4, 10, 3])
        self.assertEqual(30, result)

        result = self.run_smart_contract(engine, path, 'calculate', 5, [10, 3])
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'main', 'calculate', [5, 10, 3])
        self.assertEqual(1, result)

    def test_function_with_default_argument(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH3  # x = add(1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.CALL
            + Integer(14).to_byte_array(signed=True, min_length=1)
            + Opcode.STLOC0
            + Opcode.PUSH0
            + Opcode.PUSH6  # y = add(5, 6)
            + Opcode.PUSH5
            + Opcode.CALL
            + Integer(8).to_byte_array(signed=True, min_length=1)
            + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.LDLOC0
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.RET
            + Opcode.INITSLOT  # def add(a: int, b: int, c: int = 0)
            + b'\x00\x03'
            + Opcode.LDARG0  # return a + b + c
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.LDARG2
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('FunctionWithDefaultArgument.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([6, 11], result)

    def test_function_with_only_default_arguments(self):
        expected_output = (
            Opcode.PUSH0  # defaults
            + Opcode.PUSH0
            + Opcode.PUSH0
            + Opcode.CALL  # add()
            + Integer(20).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH0  # defaults
            + Opcode.PUSH6  # add(5, 6)
            + Opcode.PUSH5
            + Opcode.CALL
            + Integer(15).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH0  # defaults
            + Opcode.PUSH0
            + Opcode.PUSH9  # add(9)
            + Opcode.CALL
            + Integer(10).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH3  # add(1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.CALL
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.RET
            + Opcode.INITSLOT  # def add(a: int, b: int, c: int)
            + b'\x00\x03'
            + Opcode.LDARG0  # return a + b + c
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.LDARG2
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('FunctionWithOnlyDefaultArguments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([6, 9, 11, 0], result)

    def test_function_with_default_argument_between_other_args(self):
        path = self.get_contract_path('FunctionWithDefaultArgumentBetweenArgs.py')

        with self.assertRaises(SyntaxError):
            Boa3.compile(path)

    def test_call_function_with_kwarg(self):
        path = self.get_contract_path('CallFunctionWithKwarg.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(-10, result)

    def test_call_function_with_kwargs(self):
        path = self.get_contract_path('CallFunctionWithKwargs.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'positional_order')
        self.assertEqual(1234, result)

        result = self.run_smart_contract(engine, path, 'out_of_order')
        self.assertEqual(2413, result)

        result = self.run_smart_contract(engine, path, 'mixed_in_order')
        self.assertEqual(5612, result)

        result = self.run_smart_contract(engine, path, 'mixed_out_of_order')
        self.assertEqual(5621, result)

    def test_call_function_with_kwargs_with_default_values(self):
        path = self.get_contract_path('CallFunctionWithKwargsWithDefaultValues.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'positional_order')
        self.assertEqual(1234, result)

        result = self.run_smart_contract(engine, path, 'out_of_order')
        self.assertEqual(2413, result)

        result = self.run_smart_contract(engine, path, 'mixed_in_order')
        self.assertEqual(5612, result)

        result = self.run_smart_contract(engine, path, 'mixed_out_of_order')
        self.assertEqual(5621, result)

        result = self.run_smart_contract(engine, path, 'default_values')
        self.assertEqual(1034, result)

        result = self.run_smart_contract(engine, path, 'only_default_values_and_kwargs')
        self.assertEqual(204, result)

    def test_call_function_with_kwargs_only(self):
        path = self.get_contract_path('CallFunctionWithKwargsOnly.py')

        # TODO: change the test when creating a function that only accepts keywords is implemented
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_call_function_with_kwargs_self(self):
        path = self.get_contract_path('CallFunctionWithKwargsSelf.py')

        # TODO: change the test when calling a function using the class is implemented
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_call_function_with_kwargs_wrong_type(self):
        path = self.get_contract_path('CallFunctionWithKwargsWrongType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_call_function_with_kwargs_too_few_parameters(self):
        path = self.get_contract_path('CallFunctionWithKwargsTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_call_function_with_kwargs_too_many_parameters(self):
        path = self.get_contract_path('CallFunctionWithKwargsTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_call_function_with_kwargs_too_many_kw_parameters(self):
        path = self.get_contract_path('CallFunctionWithKwargsTooManyKwArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_boa2_fibonacci_test(self):
        path = self.get_contract_path('FibonacciBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 4)
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', 5)
        self.assertEqual(5, result)

        result = self.run_smart_contract(engine, path, 'main', 6)
        self.assertEqual(8, result)

        result = self.run_smart_contract(engine, path, 'main', 7)
        self.assertEqual(13, result)

        result = self.run_smart_contract(engine, path, 'main', 11)
        self.assertEqual(89, result)

    def test_boa2_method_test(self):
        path = self.get_contract_path('MethodBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1, 2)
        self.assertEqual(7, result)

        result = self.run_smart_contract(engine, path, 'main', -3, -100)
        self.assertEqual(-99, result)

    def test_boa2_method_test2(self):
        path = self.get_contract_path('MethodBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(26, result)

    def test_boa2_method_test3(self):
        path = self.get_contract_path('MethodBoa2Test3.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(13, result)

    def test_boa2_method_test4(self):
        path = self.get_contract_path('MethodBoa2Test4.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(63, result)

    def test_boa2_method_test5(self):
        path = self.get_contract_path('MethodBoa2Test5.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(15, result)

    def test_boa2_module_method_test1(self):
        path = self.get_contract_path('ModuleMethodBoa2Test1.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(True, result)

    def test_boa2_module_method_test2(self):
        path = self.get_contract_path('ModuleMethodBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(3003, result)

    def test_boa2_module_variable_test(self):
        path = self.get_contract_path('ModuleVariableBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(1260, result)

    def test_boa2_module_variable_test1(self):
        path = self.get_contract_path('ModuleVariableBoa2Test1.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(8, result)

    def test_call_void_function_with_stared_argument(self):
        path = self.get_contract_path('CallVoidFunctionWithStarredArgument.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_call_return_function_with_stared_argument(self):
        path = self.get_contract_path('CallReturnFunctionWithStarredArgument.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(sum([1, 2, 3, 4, 5, 6]), result)

    def test_return_starred_argument(self):
        path = self.get_contract_path('ReturnStarredArgumentCount.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'fun_with_starred', [1, 2, 3, 4, 5])
        self.assertEqual(5, result)

        result = self.run_smart_contract(engine, path, 'fun_with_starred', [1, 2, 3])
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', [1, 2, 3, 4, 5])
        self.assertEqual(5, result)

        result = self.run_smart_contract(engine, path, 'main', [1, 2, 3])
        self.assertEqual(3, result)

    def test_call_function_with_same_name_in_different_scopes(self):
        path = self.get_contract_path('CallFunctionsWithSameNameInDifferentScopes.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'result')
        self.assertEqual([10, 20], result)

    def test_function_with_dictionary_unpacking_operator(self):
        path = self.get_contract_path('FunctionWithDictionaryUnpackingOperator.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_functions_with_duplicated_name(self):
        path = self.get_contract_path('FunctionsWithDuplicatedName.py')
        self.assertCompilerLogs(CompilerError.DuplicatedIdentifier, path)

    def test_function_as_arg(self):
        path = self.get_contract_path('FunctionAsArg.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 123)
        self.assertEqual(123, result)

    def test_function_with_arg_as_arg(self):
        path = self.get_contract_path('FunctionWithArgAsArg.py')
        engine = TestEngine()

        list_int = [123, 456, 789]
        value = 123
        result = self.run_smart_contract(engine, path, 'main', list_int, value)
        self.assertEqual(list_int.index(value), result)

    def test_function_with_args_as_arg(self):
        path = self.get_contract_path('FunctionWithArgsAsArg.py')
        engine = TestEngine()

        list_int = [123, 456, 789]
        value = 123
        start = 0
        end = 5
        result = self.run_smart_contract(engine, path, 'main', list_int, value, start, end)
        self.assertEqual(list_int.index(value, start, end), result)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestString(BoaTest):
    default_folder: str = 'test_sc/string_test'

    SUBSTRING_NOT_FOUND_MSG = 'substring not found'

    def test_string_get_value(self):
        path = self.get_contract_path('GetValue.py')
        output = Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unit')
        self.assertEqual('u', result)
        result = self.run_smart_contract(engine, path, 'Main', '123')
        self.assertEqual('1', result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', '')

    def test_string_get_value_to_variable(self):
        path = self.get_contract_path('GetValueToVariable.py')
        output = Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unit')
        self.assertEqual('u', result)
        result = self.run_smart_contract(engine, path, 'Main', '123')
        self.assertEqual('1', result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', '')

    def test_string_set_value(self):
        path = self.get_contract_path('SetValue.py')
        self.assertCompilerLogs(CompilerError.UnresolvedOperation, path)

    def test_string_slicing(self):
        path = self.get_contract_path('StringSlicingLiteralValues.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('i', result)

    def test_string_slicing_start_larger_than_ending(self):
        path = self.get_contract_path('StringSlicingStartLargerThanEnding.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('', result)

    def test_string_slicing_with_variables(self):
        path = self.get_contract_path('StringSlicingVariableValues.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('i', result)

    def test_string_slicing_negative_start(self):
        path = self.get_contract_path('StringSlicingNegativeStart.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'unit_', result)

    def test_string_slicing_negative_end_omitted(self):
        path = self.get_contract_path('StringSlicingNegativeEndOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('test', result)

    def test_string_slicing_start_omitted(self):
        path = self.get_contract_path('StringSlicingStartOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'uni', result)

    def test_string_slicing_omitted(self):
        string_value = 'unit_test'
        byte_input = String(string_value).to_bytes()

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = 'unit_test'
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.STLOC0
            + Opcode.PUSHDATA1  # return a[:3]
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.RET
        )
        path = self.get_contract_path('StringSlicingOmitted.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('unit_test', result)

    def test_string_slicing_end_omitted(self):
        path = self.get_contract_path('StringSlicingEndOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'it_test', result)

    def test_string_slicing_with_stride(self):
        path = self.get_contract_path('StringSlicingWithStride.py')
        engine = TestEngine()

        a = 'unit_test'
        expected_result = a[2:5:2]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-6:5:2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[0:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-6:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-999:5:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[0:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-999:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[999:5:2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[0:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[999:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_values')
        self.assertEqual(expected_result, result)

    def test_string_slicing_with_negative_stride(self):
        path = self.get_contract_path('StringSlicingWithNegativeStride.py')
        engine = TestEngine()

        a = 'unit_test'
        expected_result = a[2:5:-1]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-6:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[0:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-6:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-999:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[0:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-999:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[999:5:-1]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[0:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[999:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_values')
        self.assertEqual(expected_result, result)

    def test_string_slicing_omitted_with_stride(self):
        path = self.get_contract_path('StringSlicingOmittedWithStride.py')
        engine = TestEngine()

        a = 'unit_test'
        expected_result = a[::2]
        result = self.run_smart_contract(engine, path, 'omitted_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:5:2]
        result = self.run_smart_contract(engine, path, 'omitted_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[2::2]
        result = self.run_smart_contract(engine, path, 'omitted_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-6::2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-999::2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[999::2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, result)

    def test_string_slicing_omitted_with_negative_stride(self):
        path = self.get_contract_path('StringSlicingOmittedWithNegativeStride.py')
        engine = TestEngine()

        a = 'unit_test'
        expected_result = a[::-2]
        result = self.run_smart_contract(engine, path, 'omitted_values')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:5:-2]
        result = self.run_smart_contract(engine, path, 'omitted_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[2::-2]
        result = self.run_smart_contract(engine, path, 'omitted_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-6::-2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:-1:-2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[-999::-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:-999:-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[999::-2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, result)

        a = 'unit_test'
        expected_result = a[:999:-2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, result)

    def test_string_simple_concat(self):
        path = self.get_contract_path('StringSimpleConcat.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('bye worldhi', result)

    def test_boa2_string_concat_test(self):
        path = self.get_contract_path('ConcatBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('helloworld', result)

    def test_boa2_string_concat_test2(self):
        path = self.get_contract_path('ConcatBoa2Test2.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'concat', ['hello', 'world'])
        self.assertEqual('helloworld', result)

        result = self.run_smart_contract(engine, path, 'main', 'blah', ['hello', 'world'])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 'concat', ['blah'])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 'concat', ['hello', 'world', 'third'])
        self.assertEqual('helloworld', result)

        result = self.run_smart_contract(engine, path, 'main', 'concat', ['1', 'neo'])
        self.assertEqual('1neo', result)

        result = self.run_smart_contract(engine, path, 'main', 'concat', ['', 'neo'])
        self.assertEqual('neo', result)

    def test_string_with_double_quotes(self):
        path = self.get_contract_path('StringWithDoubleQuotes.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'string_test', 'hello', 'world')
        self.assertEqual('"hell"test_symbol":world}"', result)

        result = self.run_smart_contract(engine, path, 'string_test', '1', 'neo')
        self.assertEqual('""test_symbol":neo}"', result)

        result = self.run_smart_contract(engine, path, 'string_test', 'neo', '')
        self.assertEqual('"ne"test_symbol":}"', result)

    def test_string_upper(self):
        path = self.get_contract_path('UpperStringMethod.py')
        engine = TestEngine()

        string = 'abcdefghijklmnopqrstuvwxyz'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.upper(), result)

        string = 'a1b123y3z'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.upper(), result)

        string = '!@#$%123*-/'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.upper(), result)

        string = 'áõèñ'
        result = self.run_smart_contract(engine, path, 'main', string)

        with self.assertRaises(AssertionError):
            # TODO: upper was implemented for ASCII characters only
            self.assertEqual(string.upper(), result)

    def test_string_lower(self):
        path = self.get_contract_path('LowerStringMethod.py')
        engine = TestEngine()

        string = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.lower(), result)

        string = 'A1B123Y3Z'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.lower(), result)

        string = '!@#$%123*-/'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.lower(), result)

        string = 'ÁÕÈÑ'
        result = self.run_smart_contract(engine, path, 'main', string)

        with self.assertRaises(AssertionError):
            # TODO: lower was implemented for ASCII characters only
            self.assertEqual(string.lower(), result)

    def test_string_startswith_method(self):
        path = self.get_contract_path('StartswithStringMethod.py')
        engine = TestEngine()

        string = 'unit_test'
        substring = 'unit'
        start = 0
        end = len(string)
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = 'unit'
        start = 2
        end = 6
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = 'it'
        start = 2
        end = 6
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = 'it'
        start = 2
        end = 3
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = 'unit_tes'
        start = -99
        end = -1
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = ''
        start = 0
        end = 0
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = 'unit_test'
        start = 0
        end = 99
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

        string = 'unit_test'
        substring = 'unit_test'
        start = 100
        end = 99
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.startswith(substring, start, end), result)

    def test_string_startswith_method_default_end(self):
        path = self.get_contract_path('StartswithStringMethodDefaultEnd.py')
        engine = TestEngine()

        string = 'unit_test'
        substring = 'unit'
        start = 0
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = 'unit'
        start = 2
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = 'it'
        start = 2
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = 'it'
        start = 3
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = 'unit_tes'
        start = -99
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = ''
        start = 0
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = ''
        start = 99
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

        string = 'unit_test'
        substring = 'unit_test'
        start = 0
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.startswith(substring, start), result)

    def test_string_startswith_method_defaults(self):
        path = self.get_contract_path('StartswithStringMethodDefaults.py')
        engine = TestEngine()

        string = 'unit_test'
        substring = 'unit'
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.startswith(substring), result)

        string = 'unit_test'
        substring = 'unit_test'
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.startswith(substring), result)

        string = 'unit_test'
        substring = ''
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.startswith(substring), result)

        string = 'unit_test'
        substring = '12345'
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.startswith(substring), result)

        string = 'unit_test'
        substring = 'bigger substring'
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.startswith(substring), result)

    def test_string_strip(self):
        path = self.get_contract_path('StripStringMethod.py')
        engine = TestEngine()

        string = 'abcdefghijklmnopqrstuvwxyz'
        chars = 'abcxyz'
        result = self.run_smart_contract(engine, path, 'main', string, chars)
        self.assertEqual(string.strip(chars), result)

        string = 'abcdefghijklmnopqrsvwxyz unit test abcdefghijklmnopqrsvwxyz'
        chars = 'abcdefghijklmnopqrsvwxyz '
        result = self.run_smart_contract(engine, path, 'main', string, chars)
        self.assertEqual(string.strip(chars), result)

        string = '0123456789hello world987654310'
        chars = '0987654321'
        result = self.run_smart_contract(engine, path, 'main', string, chars)
        self.assertEqual(string.strip(chars), result)

    def test_string_strip_default(self):
        path = self.get_contract_path('StripStringMethodDefault.py')
        engine = TestEngine()

        string = '     unit test    '
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.strip(), result)

        string = 'unit test    '
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.strip(), result)

        string = '    unit test'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.strip(), result)

        string = ' \t\n\r\f\vunit test \t\n\r\f\v'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.strip(), result)

    def test_isdigit_method(self):
        path = self.get_contract_path('IsdigitMethod.py')
        engine = TestEngine()

        string = '0123456789'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.isdigit(), result)

        string = '23mixed01'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.isdigit(), result)

        string = 'no digits here'
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.isdigit(), result)

        string = ''
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(string.isdigit(), result)

        string = '¹²³'
        result = self.run_smart_contract(engine, path, 'main', string)
        with self.assertRaises(AssertionError):
            # neo3-boas isdigit implementation does not verify values that are not from the ASCII
            self.assertEqual(string.isdigit(), result)

    def test_string_join_with_sequence(self):
        path = self.get_contract_path('JoinStringMethodWithSequence.py')
        engine = TestEngine()

        string = ' '
        sequence = ["Unit", "Test", "Neo3-boa"]
        result = self.run_smart_contract(engine, path, 'main', string, sequence)
        self.assertEqual(string.join(sequence), result)

        string = ' '
        sequence = []
        result = self.run_smart_contract(engine, path, 'main', string, sequence)
        self.assertEqual(string.join(sequence), result)

        string = ' '
        sequence = ["UnitTest"]
        result = self.run_smart_contract(engine, path, 'main', string, sequence)
        self.assertEqual(string.join(sequence), result)

    def test_string_join_with_dictionary(self):
        path = self.get_contract_path('JoinStringMethodWithDictionary.py')
        engine = TestEngine()

        string = ' '
        dictionary = {"Unit": 1, "Test": 2, "Neo3-boa": 3}
        result = self.run_smart_contract(engine, path, 'main', string, dictionary)
        self.assertEqual(string.join(dictionary), result)

        string = ' '
        dictionary = {}
        result = self.run_smart_contract(engine, path, 'main', string, dictionary)
        self.assertEqual(string.join(dictionary), result)

        string = ' '
        dictionary = {"UnitTest": 1}
        result = self.run_smart_contract(engine, path, 'main', string, dictionary)
        self.assertEqual(string.join(dictionary), result)

    def test_string_index(self):
        path = self.get_contract_path('IndexString.py')
        engine = TestEngine()

        string = 'unit test'
        substring = 'i'
        start = 0
        end = 4
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.index(substring, start, end), result)

        string = 'unit test'
        substring = 'i'
        start = 2
        end = 4
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.index(substring, start, end), result)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSTRING_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 3, 4)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSTRING_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 4, -1)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSTRING_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 0, -99)

        string = 'unit test'
        substring = 'i'
        start = 0
        end = -1
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.index(substring, start, end), result)

        string = 'unit test'
        substring = 'n'
        start = 0
        end = 99
        result = self.run_smart_contract(engine, path, 'main', string, substring, start, end)
        self.assertEqual(string.index(substring, start, end), result)

    def test_string_index_end_default(self):
        path = self.get_contract_path('IndexStringEndDefault.py')
        engine = TestEngine()

        string = 'unit test'
        substring = 't'
        start = 0
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.index(substring, start), result)

        string = 'unit test'
        substring = 't'
        start = 4
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.index(substring, start), result)

        string = 'unit test'
        substring = 't'
        start = 6
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.index(substring, start), result)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSTRING_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 'i', 99)

        with self.assertRaisesRegex(TestExecutionException, f'{self.SUBSTRING_NOT_FOUND_MSG}$'):
            self.run_smart_contract(engine, path, 'main', 'unit test', 't', -1)

        string = 'unit test'
        substring = 'i'
        start = -10
        result = self.run_smart_contract(engine, path, 'main', string, substring, start)
        self.assertEqual(string.index(substring, start), result)

    def test_string_index_defaults(self):
        path = self.get_contract_path('IndexStringDefaults.py')
        engine = TestEngine()

        string = 'unit test'
        substring = 'u'
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.index(substring), result)

        string = 'unit test'
        substring = 't'
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.index(substring), result)

        string = 'unit test'
        substring = ' '
        result = self.run_smart_contract(engine, path, 'main', string, substring)
        self.assertEqual(string.index(substring), result)

    def test_string_index_mismatched_type(self):
        path = self.get_contract_path('IndexStringMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)
